{$A+,B-,F+,G+,N+,E-,P+,Q-,R-,T-,V+,X+}
{****************************************************************************
 * SoundLib 2 MODule file playback system -                                 *
 * copyright (c) 1997-2001 by Stefan Goehler                                *
 * Copyright agreement in soundlib.txt - illegal copy without this file!    *
 * Version: 1a3                                                             *
 ****************************************************************************}
{

History
  a2:
  .S3M supported
  .fixed some minor MOD playback bugs
  .fixed setpos bug, changed setpos function
  .added getpos function
  .changed mod panning settings (20% Stereo mix)
  .added note retrigger effect, including S3M extension
  .modified setpos for backward scrolling part to work faster
  .more MOD formats supported
  .global volume support added (for S3M and XM effects)
  .S3M tremor effect added
  .XM support added (better don't try it yet ;)

  a3
  .fixed bug with pattern jumping (false pos if command more than once on line)
  .XM calculation routines extended, XM has got its own line calculation
   procedure now
  .extended by FPC support
  .XM loader should load now all XMs
  .new position calculation - works now correct with XM
  .XM sample mapping works (phew, MUCH work!!)
  .simple volume ramping included to avoid clicking
}

{
Known issues:
 -XM support is included, but due to too few and too lousy doc it isn't
  finished yet (need some docs about those envelopes)

}

unit snd_mod;

interface
{$IFNDEF FPC}
uses objects,sndlib2,snd_chn,stream32;
{$ELSE}
uses objects,sndlib2,snd_chn;
{$ASMMODE INTEL}
{$ENDIF}

type
  { MOD file structures }
  tmodhdr = record
    songname : array[0..19] of char;

    sampleinfo : array[1..31] of record
      name       : array[0..21] of char;
      length     : word;
      finetune   : byte;
      volume     : byte;
      loopstart  : word;
      looplength : word;
    end;

    songlength : byte;
    ciaaspeed  : byte;
    arrange    : array[0..127] of byte;
    id         : array[0..3] of char;
  end;

  { S3M file structures }
  ts3mhdr = record
    songname    : array[0..27] of char;
    hex1Ah      : byte;
    typ         : byte;
    reserved1   : array[0..1] of byte;
    songlength  : word;
    instruments : word;
    patterns    : word;
    flags       : word;
    version     : word;
    fileformat  : word;
    id          : array[0..3] of char;
    globalvol   : byte;
    starttempo  : byte;
    startbpm    : byte;
    mastervol   : byte;
    ultaclick   : byte;
    defpan      : byte;
    reserved2   : array[0..7] of byte;
    special     : word;
    channelinfo : array[0..31] of byte;
  end;

  ts3minstr = record
    typ       : byte;
    filename  : array[0..11] of char;
    unused    : byte;
    segment   : word;
    size      : longint;
    loopstart : longint;
    loopend   : longint;
    volume    : byte;
    disk      : byte;
    packing   : byte;
    flags     : byte;
    c2spd     : longint;
    reserved  : array[0..3] of byte;
    GrvsRamPos: word;{Internal use for GUS}
    loopexpans: word;{Internal use for SB}
    lastpos   : longint;{Internal use for SB}
    samplename: array[0..27] of char;
    kennung   : array[0..3] of char;{SCRS}
  end;

  { XM file structures }
  txmstart = record { the few first bytes of a XM }
    id         : array[0..16] of char; { 'Extended module: ' }
    songname   : array[0..19] of char;
    hex1Ah     : byte;
    tracker    : array[0..19] of char;
    version    : word;                 { Current version is $0103}
  end;

  txmhdr = record { follows the XM start }
    headersize   : longint;
    songlength   : word; { Song length in pattern order table  }

    restart      : word; { Loop position                       }
    channels     : word; { must be multiple of 2               }
    patterns     : word; { number of patterns; max 256         }
    instruments  : word; { number of instruments; max 128      }
    flags        : word; { only bit 0 used for frequency table }
    starttempo   : word;
    startbpm     : word;
  end;


  tpatt = record
    per : word;
    ins : byte;
    vol : byte;
    cmd : byte;
    arg : byte;
  end;

  pxmsmpmapping = ^txmsmpmapping;
  txmsmpmapping = array[1..128,1..96] of byte;

  pmodule = ^tmodule;
  tmodule = object(tsound)
    actline      : longint;
    actpattern   : longint;
    tempo        : word;
    bpm          : word;
    tick         : word;
    arrange      : array[0..255] of byte;
    arrangespeed : array[0..255] of record
                     tempo,
                     bpm  : byte;
                   end;
    pattern      : array[0..255] of record
                     p    : pointer;
                     size : word;
                     rows : byte;
                   end;
    instruments  : word;
    patterns     : word;
    linesize     : word;
    cs           : tchannelset;
    playchnls    : word;
    flags        : longint;
    songlength   : word;
    songname     : string[30];
    oscillator   : longint;
    pershift     : byte;
    t0volslide   : boolean;
    format       : word;

    constructor init;
    destructor  done;virtual;
    procedure   play;virtual;
    procedure   stop;virtual;
    procedure   pause;virtual;
    function    isplaying : boolean;virtual;
    function    ispaused : boolean;virtual;

    function    setpos(pos : longint) : boolean;virtual;
    function    getpos : longint;virtual;
    function    getsize : longint;virtual;

    function    getversionstring : string;virtual;

    function    change : boolean;virtual;
    { Out of interest for you: }
    procedure   callproc;virtual;
    private
    busy         : boolean;
    lastpos      : longint;
    patdelay     : word;   { Pattern delay value }
    minslide,
    maxslide     : word;   { Porta slide min/max }

    procedure   calcline;virtual;
    procedure   calcdata_tickn0;
    procedure   calcdata_tick0;
    procedure   calcdata_tick;
  end;

  pmod = ^tmod;
  tmod = object(tmodule)
    header       : tmodhdr;

    constructor init(var dest : tsounddriver;var f : tstream);

    procedure   play;virtual;
  end;

  ps3m = ^ts3m;
  ts3m = object(tmodule)
    header  : ts3mhdr;
    mapping : array[0..31] of byte;
    panlist : array[0..31] of byte;

    constructor init(var dest : tsounddriver;var f : tstream);

    procedure   play;virtual;
  end;

  pxm = ^txm;
  txm = object(tmodule)
    header  : txmhdr;

    constructor init(var dest : tsounddriver;var f : tstream);
    destructor done;virtual;

    procedure   play;virtual;

    procedure   calcline;virtual;

    private
    startinfo  : txmstart;
    mapping    : array[1..128,0..15] of byte;
    smpmapping : pxmsmpmapping;
  end;

const
{$IFNDEF FPC}
  tmod_versionstr    = 'MODule player 1a3.0210.BP';
{$ELSE}
  tmod_versionstr    = 'MODule player 1a3.0210.FPC';
{$ENDIF}
  tmod_versionnum    = $0100;
  tmod_versionsubnum = $0210;

  fmt_mod = $0001;
  fmt_s3m = $0002;
  fmt_xm  = $0003;

implementation
uses snd_cnv
{$IFNDEF FPC}
 ,newfrontier
{$ENDIF}
 ;

function makestring(var data;size : word) : string;
var
  i : byte;
begin
  {$IFNDEF FPC}
  for i := 1 to size do makestring[i] := chr(mem[seg(data):ofs(data)+i-1]);
  {$ELSE}
  for i := 1 to size do makestring[i] := char(pointer(@data+i-1)^);
  {$ENDIF}
  makestring[0] := chr(size);
  makestring[size+1] := #0;
end;

function amiword(w : word): word;assembler;
asm
  mov   ax,w
  xchg  ah,al
end;

constructor tmodule.init;
begin
end;

destructor tmodule.done;
var
  i : integer;
begin
  flags := flags and not snd_play;
  for i := 0 to patterns-1 do freemem(pattern[i].p,pattern[i].size);
  for i := 0 to playchnls-1 do cs.channel[i].sound.done;
  cs.close;
  inherited done;
end;

procedure tmodule.play;
begin
end;

procedure tmodule.stop;
begin
  flags := flags and not (snd_play or snd_pause);
end;

procedure tmodule.pause;
var
  i : integer;

begin
  busy := true;
  if (flags and snd_pause <> 0) then
  begin
    flags := (flags and not snd_pause) or snd_play;
    for i := 0 to playchnls-1 do cs.channel[i].sound.pause;
  end
  else if (flags and snd_play <> 0) then
  begin
    flags := (flags and not snd_play) or snd_pause;
    for i := 0 to playchnls-1 do cs.channel[i].sound.pause;
  end;
  busy := false;
end;

function tmodule.isplaying : boolean;
begin
  isplaying := flags and snd_play <> 0;
end;

function tmodule.ispaused : boolean;
begin
  ispaused := flags and snd_pause <> 0;
end;

function tmodule.setpos(pos : longint) : boolean;
var
  i          : integer;
  difference : longint;

  function getpattern(pos : longint) : longint;
  { Returns pattern of pos value }
  var
    count : longint;

  begin
    count := 0;
    while (pos > word(pattern[arrange[count]].rows)+1) do
    begin
      dec(pos,word(pattern[arrange[count]].rows)+1);
      inc(count);
    end;
    getpattern := count+1;
  end;

  function getline(pos : longint) : longint;
  { Returns line of pos value }
  var
    count : longint;

  begin
    count := 0;
    while (pos > word(pattern[arrange[count]].rows)+1) do
    begin
      dec(pos,word(pattern[arrange[count]].rows)+1);
      inc(count);
    end;
    getline := pos;
  end;

  function getposition(patt,line : longint) : longint;
  { Returns pos value by calculating pattern and line }
  var
    count,i : longint;

  begin
    count := 0;
    for i := 0 to patt do inc(count,word(pattern[arrange[i]].rows)+1);
    getposition := count+line;
  end;

begin
  setpos := false;
  if (flags and snd_pause <> 0) then exit;
  { Fail if pattern > max pattern }
  if (getpattern(pos)-1 > songlength) then exit;
  if (pos < 0) then
  begin
    for i := 0 to playchnls-1 do cs.channel[i].sound.stop;
    pos := 0;
    actpattern := 1;
    actline    := 0;
    setpos     := true;
    exit;
  end;

  { Set busy flag - otherwise machine may hang! }
  busy := true;

  difference := pos-getposition(actpattern,actline);

  if (difference < 0) then
  begin
    { Backwards: stop all channels! }
    for i := 0 to playchnls-1 do cs.channel[i].sound.stop;
    actpattern := getpattern(pos)-1;
    actline    := 0;
    for i := 0 to getline(pos)-1 do
    begin
      calcline;
      if (actline >= getline(pos)) then break;
    end;
  end
  else { Forward }
  begin
    for i := 0 to difference-1 do
    begin
      calcline;
      if (actpattern >= getpattern(pos))and(actline >= getline(pos)) then break;
    end;
  end;
  setpos := true;
  busy   := false;
end;

function tmodule.getpos : longint;

  function getposition(patt,line : longint) : longint;
  { Returns pos value by calculating pattern and line }
  var
    count,i : longint;

  begin
    count := 0;
    for i := 0 to patt do inc(count,word(pattern[arrange[i]].rows)+1);
    getposition := count+line;
  end;

begin
  getpos := getposition(actpattern-1,actline);
end;

function tmodule.getsize : longint;

  function getposition(patt,line : longint) : longint;
  { Returns pos value by calculating pattern and line }
  var
    count,i : longint;

  begin
    count := 0;
    for i := 0 to patt do inc(count,word(pattern[arrange[i]].rows)+1);
    getposition := count+line;
  end;

begin
  getsize := getposition(songlength-2,pattern[arrange[songlength-1]].rows);
end;

function tmodule.getversionstring : string;
begin
  getversionstring := tmod_versionstr;
end;

function tmodule.change : boolean;
var
  actpos : longint;

begin
  actpos := actpattern*64*32+actline*32+tick;
  change := (actpos <> lastpos);
  lastpos := actpos;
end;

procedure tmodule.calcline;
type
  tln = array[0..31] of tpatt;

var
  ln        : ^tln;
  w         : word;
  ch,i      : integer;
  dontplay  : boolean;
  curr      : tpatt;
  per       : longint;
  last      : word;
  broken    : boolean;

begin
  if (actline = word(pattern[arrange[actpattern]].rows)+1) then
  begin
    actline := 0;
    inc(actpattern);
    arrangespeed[actpattern].bpm   := bpm;
    arrangespeed[actpattern].tempo := tempo;
  end;

  if (actpattern >= songlength) then
  begin
    flags := flags and not snd_play;
    exit;
  end;

  {$IFNDEF FPC}
  ln := ptr(seg(pattern[arrange[actpattern]].p^),
            ofs(pattern[arrange[actpattern]].p^)+actline*linesize);
  {$ELSE}
  ln := pointer(pattern[arrange[actpattern]].p+actline*linesize);
  {$ENDIF}

  broken := false; { Clear pattern not broken flag }

  for ch := 0 to playchnls-1 do with cs.channel[ch] do
  begin
    curr := ln^[ch];
    per  := curr.per;

    { Ignore porta to note if instrument different }
    if (curr.cmd = $03)and(curr.ins <> 0)and(curr.ins <> voice) then
    begin
      curr.cmd := 0;
      curr.arg := 0;
    end;

    if (per <> 0)and(per <> mdkeyoff) then
    begin
      w := curr.ins;
      if (w = 0) then w := voice;
      if (cs.instrument[w].tuning <> 8363)and(cs.instrument[w].tuning <> 0) then
          per := 8363 * per div cs.instrument[w].tuning;

      if (curr.cmd <> $03)and(per <> 0) then
      begin
        period := per;
        freq   := oscillator div per;
      end;
    end;

    { Don't play note if note delay, porta to note or key off }
{    dontplay := ((curr.cmd = $0E)and((curr.arg shr 4) = $0D))or
                 ((curr.cmd = $03)and(curr.ins = 0))or
                 (curr.per = ntkeyoff);}
    case format of
      fmt_mod : dontplay := ((curr.cmd = $0E)and((curr.arg shr 4) = $0D))or
                            ((curr.cmd = $03)and(curr.ins = 0))or
                            (curr.per = mdkeyoff);
      fmt_s3m : dontplay := ((curr.cmd = $0E)and((curr.arg shr 4) = $0D))or
                            (curr.cmd = $03)or
                            (curr.per = mdkeyoff);
    end;

    if (not dontplay) then
    if (curr.ins <> 0){and((per <> 0)or(curr.cmd = 0))} then
      cs.play(ch,curr.ins,freq,cs.instrument[curr.ins].volume) else
              if (per <> 0) then
              cs.play(ch,voice,freq,volume)
              else if (curr.ins <> 0)and(per = 0) then
              cs.play(ch,voice,freq,volume);

    { Retrigger volume if porta note and same instrument on line }
    if (curr.cmd = $03)and(curr.ins <> 0)and(curr.ins = voice)and
       (format <> fmt_mod)and(curr.vol = 0)
       then curr.vol := cs.instrument[voice].volume+$10;

    if ((arpeggio)or(vibrato))and(period <> 0) then
    begin
      cs.setfrequency(ch,freq);
      arpeggio    := false;
      vibrato     := false;
    end;

    { Reset all FX which are linewise }
    tremolo     := false;
    porta       := false;
    bfinepslide := false;
    bfinevslide := false;
    vslide      := false;
    bportaslide := false;
    retrig      := 0;
    tremor      := false;
    last        := lasteffect;
    lasteffect  := curr.cmd+curr.arg shl 8;

    if (curr.vol <> 0)and(curr.vol <= $50) then
    begin
      volume := curr.vol-$10;
      if ((curr.cmd = $0E)and((curr.arg shr 4) = $0D))
      then else cs.setvolume(ch,volume);
    end;

    if (curr.per = mdkeyoff) then
    begin
      volume := 0;
      cs.setvolume(ch,volume);
    end;

    case curr.cmd of
      $00 : if (curr.arg <> 0) then
            begin{ok; Arpeggio          }
              arpeggio  := true;
              arpeggiox := (curr.arg shr 4);
              arpeggioy := (curr.arg and $F);
            end else lasteffect := mdkeyoff;
      $01 : begin{ok; Portamento down   }
              portaslide  := -(integer(curr.arg) shl pershift);
              bportaslide := true;
            end;
      $02 : begin{ok; Portamento up     }
              portaslide  := integer(curr.arg) shl pershift;
              bportaslide := true;
            end;
      $03 : begin{ok; Porta to note     }
              if (per <> 0) then portanote := per;
              if (curr.arg <> 0) then
              begin
                portaspeed := word(curr.arg) shl pershift;
                if (period > portanote) then portaspeed := -portaspeed;
              end;
              porta := (portanote <> 0);
            end;
      $04 : begin{ok; Set vibrato       }
              if (curr.arg and $F0 <> 0) then vibratospeed := curr.arg shr 4;
              if (curr.arg and $F <> 0) then vibratodepth := (curr.arg and $F) shl pershift;
              vibrato := true;
            end;
      $05 : begin{ok; Porta+vol slide   }
              if (curr.arg and $F <> 0) then volumeslide := -(curr.arg and $F)
                else if (curr.arg and $F0 <> 0) then volumeslide := curr.arg shr 4;
              porta := (portanote <> 0);
              vslide := true;
            end;
      $06 : begin{ok; Vibrato+vol slide }
              if (curr.arg and $F <> 0) then volumeslide := -(curr.arg and $F)
              else if (curr.arg and $F0 <> 0) then volumeslide := curr.arg shr 4;
              vibrato := true;
              vslide  := true;
            end;
      $07 : begin{ok; Tremelo           }
              if (curr.arg and $F0 <> 0) then tremolospeed := curr.arg shr 4;
              if (curr.arg and $F <> 0) then tremolodepth := curr.arg and $F;
              tremolo := true;
            end;
      $08 : begin{ok; Panning           }
              panning := integer(curr.arg)*2-256;
              cs.setvolume(ch,cs.channel[ch].volume);
            end;
      $09 : begin{ok; Set instrument pos}
              if (curr.arg <> 0) then lastsetpos := longint(curr.arg) shl 8;
              cs.setpos(ch,lastsetpos);
            end;
      $0A : begin
              if (format = fmt_s3m) then
              begin
                if (curr.arg and $F <> 0) then volumeslide := -(curr.arg and $F)
                else if (curr.arg and $F0 <> 0) then volumeslide := curr.arg shr 4;
              end else
              begin
                if (curr.arg and $F <> 0) then volumeslide := -(curr.arg and $F)
                else volumeslide := curr.arg shr 4;
              end;
              vslide  := true;
            end;
      $0B : if (curr.arg <= patterns) then
            begin{ok}
              actline    := 63;
              actpattern := curr.arg-1;
            end;
      $0C : begin{ok; Set channel volume }
              if (curr.arg > 64) then curr.arg := 64;
              cs.channel[ch].volume := curr.arg;
              cs.setvolume(ch,curr.arg);
            end;
      $0D : if (not broken) then
            begin{ok; Pattern break      }
              inc(actpattern);
              actline := 10*(curr.arg shr 4)+curr.arg and $0F-1;
              if (actline > 62) then actline := 62;
              broken := true;
            end;
      $0E : case (curr.arg shr 4) of
              $01 : begin { ok; Fine portamento up   }
                      finepslide := -((curr.arg and $F) shl pershift);
                      bfinepslide := (finepslide <> 0);
                    end;
              $02 : begin { ok; Fine portamento down }
                      finepslide := (curr.arg and $F) shl pershift;
                      bfinepslide := (finepslide <> 0);
                    end;
              $04 : begin { Set vibrato waveform }
                      vibratotype := curr.arg and 7;
                      if (vibratotype and 3) = 3 then
                        vibratotype := random(2)+curr.arg and 4;
                    end;
              $05 : begin { Set finetune }
                      cs.instrument[voice].tuning := getc2spd(curr.arg and $F);
                    end;
              $07 : begin { Set tremolo waveform }
                      tremolotype := curr.arg and 7;
                      if (tremolotype and 3) = 3 then
                        tremolotype := random(2)+curr.arg and 4;
                    end;
              $08 : begin { ok; Position panning }
                      panning := (curr.arg and $F);
                      if (panning > 7) then inc(panning);
                      panning := panning*32-256;
                      cs.setvolume(ch,cs.channel[ch].volume);
                    end;
              $09 : begin { ok; Retrigger note   }
                      retrig     := (curr.arg and $F);
                      retrigtype := 0;
                    end;
              $0A : begin { ok; Fine volume slide up   }
                      finevslide := curr.arg and $F;
                      bfinevslide := (finevslide <> 0);
                    end;
              $0B : begin { ok; Fine volume slide down }
                      finevslide := -integer(curr.arg and $F);
                      bfinevslide := (finevslide <> 0);
                    end;
              $0C : begin { ok; Cut note }
                      if (curr.arg and $F <> 0) then volcutcount := (curr.arg and $F);
                    end;
              $0D : begin { ok; Note delay }
                      notedelay := (curr.arg and $F);
                      if (curr.ins <> 0) then voice := curr.ins;
                      if (curr.vol = 0) then volume := cs.instrument[voice].volume;
                    end;
              $0E : begin { ok; Pattern delay }
                      patdelay := (curr.arg and $F);
                    end;
            end;
      $0F : if (curr.arg <> 0) then
            begin{ ok; Set bpm/tempo }
              if (curr.arg < 32) then tempo := curr.arg
              else
              begin
                bpm := curr.arg;
                setbuffersize(round(dat.mainfreq/(bpm*2/5)));
              end;
            end;
      $10 : begin { Set global volume }
              if (curr.arg > 64) then curr.arg := 64;
              cs.globalvol := curr.arg*4;
              for i := 0 to playchnls-1 do cs.setvolume(i,cs.channel[i].volume);
            end;
      $1B : begin
              retrig     := (curr.arg and $F);
              retrigtype := (curr.arg shr 4);
            end;
      $1D : begin { Tremor }
              tremor      := true;
              tremorcount := 0;
              tremorx     := curr.arg shr 4+1;
              tremory     := tremorx+curr.arg and 7+1;
            end;
      $21 : case (curr.arg shr 4) of
              $01 :  begin { ok; Extra fine portamento up   }
                      finepslide := -(curr.arg and $F);
                      bfinepslide := (finepslide <> 0);
                    end;
              $02 : begin { ok; Extra Fine portamento down }
                      finepslide := curr.arg and $F;
                      bfinepslide := (finepslide <> 0);
                    end;
            end;
      $F4 : begin{ ok; S3M Fine vibrato    }
              if (curr.arg and $F0 <> 0) then vibratospeed := curr.arg shr 4;
              if (curr.arg and $F <> 0) then vibratodepth := curr.arg and $F;
              vibrato := true;
            end;
      $F5 : if (curr.arg <> 0) then tempo := curr.arg; { Set S3M tempo }
      else lasteffect := mdnotavail;
    end;
  end;
  inc(actline);
end;

procedure tmodule.calcdata_tick0;
var
  i,tp  : integer;
  delta : longint;

begin
  for i := 0 to playchnls-1 do with cs.channel[i] do
  if (voice <> 0) then
  begin

    if (bfinevslide) then
    begin
      inc(volume,finevslide);
      if (volume < 0) then
      begin
        bfinevslide := false;
        volume      := 0;
      end else
      if (volume >= 64) then
      begin
        bfinevslide := false;
        volume      := 64;
      end;
      cs.setvolume(i,volume);
    end;

    if (bfinepslide)and(period <> 0) then
    begin
      inc(period,finepslide);
      if (period >= maxslide) then
      begin
        period      := maxslide;
        bfinepslide := false;
      end else
      if (period <= minslide) then
      begin
        period      := minslide;
        bfinepslide := false;
      end;
      freq := oscillator div period;
      cs.setfrequency(i,freq);
    end;
  end;
end;

procedure tmodule.calcdata_tick;
var
  i,vol : integer;
  delta : longint;

begin
  for i := 0 to playchnls-1 do with cs.channel[i] do
  begin
    if (vfading) then
    begin
      dec(fadeoutvol,volfade);
      if (fadeoutvol <= 0) then
      begin
        volume  := 0;
        fadeoutvol := 65536;
        cs.setvolume(i,volume);
        vfading := false;
      end;
    end;

    if (retrig <> 0)and(tick mod retrig = 0) then
    begin
      vol := cs.channel[i].volume;
      case retrigtype of
        $0,$8 :;
        $1    : dec(vol,1);
        $2    : dec(vol,2);
        $3    : dec(vol,4);
        $4    : dec(vol,8);
        $5    : dec(vol,16);
        $6    : vol := vol*2 div 3;
        $7    : vol := vol shr 1;
        $9    : inc(vol,1);
        $A    : inc(vol,2);
        $B    : inc(vol,4);
        $C    : inc(vol,8);
        $D    : inc(vol,16);
        $E    : vol := vol*3 div 2;
        $F    : vol := vol shl 1;
      end;
      if (vol < 0) then vol := 0
      else if (vol > 64) then vol := 64;
      cs.channel[i].volume := vol;
      if (voice <> 0) then cs.play(i,voice,freq,cs.channel[i].volume);
    end;

    if (vslide)and((tick <> 0)or(t0volslide)) then
    begin
      inc(volume,volumeslide);
      if (volume <= 0) then
      begin
        vslide := false;
        volume := 0;
      end else
      if (volume >= 64) then
      begin
        volumeslide := 0;
        volume      := 64;
      end;
      cs.setvolume(i,volume);
    end;

    if (tremor)and((tremorcount mod tremorx = 0)or(tremorcount mod tremory = 0)) then
    begin
      if (tremorcount mod tremorx = 0) then
        cs.setvolume(i,cs.channel[i].volume)
        else cs.setvolume(i,0);
    end;
    inc(tremorcount);
  end;
end;

procedure tmodule.calcdata_tickn0;
var
  i,vol : integer;
  delta : longint;

begin
  for i := 0 to playchnls-1 do with cs.channel[i] do
  begin
    if (notedelay <> 0)and(tick = notedelay) then
    begin
      notedelay := 0;
      if (voice <> 0) then cs.play(i,voice,freq,cs.channel[i].volume);
    end;

    if (bportaslide)and(period <> 0) then
    begin
      inc(period,portaslide);
      if (period >= maxslide) then
      begin
        period      := maxslide;
        bportaslide := false;
      end else
      if (period <= minslide) then
      begin
        period      := minslide;
        bportaslide := false;
      end;
      freq := oscillator div period;
      cs.setfrequency(i,freq);
    end;

    if (porta)and(period <> 0) then
    begin
      inc(period,portaspeed);
      if (portaspeed < 0)and(period < portanote) then
      begin
        period := portanote;
        porta  := false;
      end else
      if (portaspeed > 0)and(period > portanote) then
      begin
        period := portanote;
        porta  := false;
      end;
      freq := oscillator div period;
      cs.setfrequency(i,freq);
    end;

    if (vibrato)and(period <> 0) then
    begin
      case vibratotype and 3 of
        0 : delta := (vibrato_sine[vibratopos]*vibratodepth);
        1 : begin
              delta := (vibratopos and 31)*8;
              if (vibratopos > 31) then delta := 255-delta;
            end;
        2 : if (vibratopos < 32) then delta := 255 else delta := -255
      end;

      delta := delta div 128;
      cs.setfrequency(i,oscillator div (period+delta));
      vibratopos := (vibratopos+vibratospeed) and 63;
    end;

    if (tremolo) then
    begin
      case tremolotype and 3 of
        0 : delta := (vibrato_sine[tremolopos]*tremolodepth);
        1 : begin
              delta := (tremolopos and 31)*8;
              if (tremolopos > 31) then delta := 255-delta;
            end;
        2 : if (tremolopos < 32) then delta := 255 else delta := -255
      end;

      delta := volume+(delta div 64);
      if (volume < 0) then volume := 0;
      if (volume > 64) then volume := 64;
      cs.setvolume(i,volume+delta);
      tremolopos := (tremolopos+tremolospeed) and 63;
    end;

    if (arpeggio)and(period <> 0) then
    begin
      case tick mod 3 of
        0 : cs.setfrequency(i,freq);
        1 : cs.setfrequency(i,oscillator div getnewhalftone(period,arpeggiox));
        2 : cs.setfrequency(i,oscillator div getnewhalftone(period,arpeggioy));
      end;
    end;

    if (volcutcount > 0) then
    begin
      dec(volcutcount);
      if (volcutcount = 0) then
      begin
        cs.channel[i].volume := 0;
        cs.setvolume(i,0);
      end;
    end;
  end;
end;

procedure tmodule.callproc;
var
  i : integer;

begin
  if (flags and snd_play <> 0)and(not busy) then
  begin
    tick := (tick+1) mod tempo;
    if (tick = 0) then
    begin
      if (patdelay > 0) then dec(patdelay) else calcline;
      calcdata_tick0;
    end else
      calcdata_tickn0;
      calcdata_tick;
  end;
end;

constructor tmod.init(var dest : tsounddriver;var f : tstream);
type
  tln = array[0..31] of tpatt;

var
  ln       : ^tln;

  s        : string;
  i,j,line : integer;
  w        : word;
  p        : dataptr;
  l        : longint;
  pt       : pointer;

begin
  if (f.status <> stok) then fail;

  tsound.init(dest,0);

  f.read(header,1084);
  s := header.id;
  s[0] := #4;

  { Try to detect module format and set channels }
  if (s = 'TDZ1') then playchnls := 1 else
  if (s = 'TDZ2') then playchnls := 2 else
  if (s = 'TDZ3') then playchnls := 3 else
  if (s = 'M.K.')or(s = 'M!K!')or(s = 'FLT4') then
    playchnls := 4 else
  if (s = 'OCTA')or(s = 'CD81')or(s = 'FLT8') then
    playchnls := 8 else
  if (pos('CHN',s) <> 0) then
  begin
    delete(s,2,3);
    val(s,i,j);
    if (j <> 0) then
    begin
      tsound.done;
      fail;
    end;
    playchnls := i;
  end else
  if (pos('CH',s) <> 0) then
  begin
    delete(s,3,2);
    val(s,i,j);
    if (j <> 0) then
    begin
      tsound.done;
      fail;
    end;
    playchnls := i;
  end else
  begin
    tsound.done;
    fail;
  end;

  { Initialize sound object, set default of 50 bpm }
  if (playchnls > 4) then tsound.setchannels(playchnls shr 1)
  else tsound.setchannels(playchnls);
  setbuffersize(dat.mainfreq div 50);

  { Set format restrictions }
  format      := fmt_mod;
  oscillator  := osc_freq;
  pershift    := 0;
  minslide    := 113;
  maxslide    := 856;
  linesize    := playchnls*sizeof(tpatt);
  instruments := 31;

  arrangespeed[0].bpm   := 125;
  arrangespeed[0].tempo := 6;

  for i := 0 to 19 do
  if (header.songname[i] >= #32) then
    songname[i+1] := header.songname[i] else songname[i+1] := #32;
  songname[0] := #20;

  cs.init;
  for i := 0 to maxchnls-1 do cs.channel[i].panning := channelval[i]*512-256;
  for i := 0 to maxchnls-1 do if (cs.channel[i].panning < 0) then
    inc(cs.channel[i].panning,stereomix) else dec(cs.channel[i].panning,stereomix);

  for i := 0 to playchnls-1 do cs.channel[i].sound.init(self);

  patterns := 0;
  for i := 0 to 127 do if (header.arrange[i] > patterns) then
    patterns := header.arrange[i];
  inc(patterns);

  songlength := header.songlength;
  move(header.arrange,arrange,128);


  { Load and convert all patterns }
  getmem(pt,playchnls*4*64);

  for i := 0 to patterns-1 do
  begin
    with pattern[i] do
    begin
      size := linesize*64;
      rows := 63;
      getmem(p,size);
    end;

    f.read(pt^,playchnls*4*64);

    for line := 0 to 63 do
    begin
      {$IFNDEF FPC}
      ln := ptr(seg(pattern[i].p^),ofs(pattern[i].p^)+line*linesize);
      {$ELSE}
      ln := pointer(pattern[i].p+line*linesize);
      {$ENDIF}
      for j := 0 to playchnls-1 do with ln^[j] do
      begin
        {$IFNDEF FPC}
        l   := meml[seg(pt^):ofs(pt^)+line*playchnls*4+j*4];
        {$ELSE}
        l   := longint(pointer(pt+line*playchnls*4+j*4)^);
        {$ENDIF}

        ins := (l and $10)+(l shr 20) and $F;
        per := amiword(l) and $0FFF;
        cmd := (l shr 16) and $F;
        arg := (l shr 24);
        vol := 0;
      end;
    end;
  end;

  freemem(pt,playchnls*4*64);

  { Load instrument settings and names }
  for i := 1 to instruments do with cs.instrument[i] do
  begin
    for j := 0 to 21 do
    if (header.sampleinfo[i].name[j] >= #32) then
      name[j+1] := header.sampleinfo[i].name[j] else name[j+1] := #32;
    name[0]   := #22;
    size      := longint(amiword(header.sampleinfo[i].length)) shl 1;
    loopstart := longint(amiword(header.sampleinfo[i].loopstart)) shl 1;
    l := longint(amiword(header.sampleinfo[i].looplength)) shl 1;
    if (l = 2) then l := 0;
    loopend   := loopstart+l;
    if (loopend = loopstart) then
    begin
      loopend   := 0;
      loopstart := 0;
    end;
    volfade := 65536;
    volume := header.sampleinfo[i].volume;
    tuning  := getc2spd(header.sampleinfo[i].finetune and $F);
    if (loopend > size)or(loopend = 2) then loopend := 0;
    if (size = 2) then size := 0;
    ofreq := stdfreq;
  end;

  { Load and convert instrument samples }
  for i := 1 to instruments do with cs.instrument[i] do
  if (size > 0) then
  with cs.instrument[i] do
  begin
    {$IFNDEF FPC}
    getmem32(s,size shl 1);
    getmem32(p,size);
    f.read32(ptr48(p,0),size);
    {$ELSE}
    getmem(s,size shl 1);
    getmem(p,size);
    f.read(p^,size);
    {$ENDIF}
    if (f.status <> stok) then
    begin
      {$IFNDEF FPC}
      freemem32(s);
      freemem32(p);
      {$ELSE}
      freemem(s);
      freemem(p);
      {$ENDIF}
      size := 0;
      open := false;
    end
    else open      := true;
    {$IFNDEF FPC}
    aud_s8_to_s16(ptr48(p,0),ptr48(s,0),size);
    freemem32(p);
    {$ELSE}
    aud_s8_to_s16(p,s,size);
    freemem(p);
    {$ENDIF}
  end;
  cs.ramp; { Apply volume ramping }
  initialized := true;
end;

procedure tmod.play;
var
  i : integer;

begin
  { If paused, unpause }
  if (flags and snd_pause <> 0) then
  begin
    flags := (flags and not snd_pause) or snd_play;
    for i := 0 to playchnls-1 do cs.channel[i].sound.pause;
  end
  else { else restart from beginning }
  begin
    actline    := 0;
    actpattern := 0;
    flags      := flags or snd_play;
    tick       := 0;
    tempo      := 6;
    bpm        := 125;
  end;
end;

constructor ts3m.init(var dest : tsounddriver;var f : tstream);
type
  tln = array[0..31] of tpatt;

  ts3mpatt = record
    o_n : byte;
    ins : byte;
    vol : byte;
    cmd : byte;
    arg : byte;
  end;
  tba = array[0..0] of byte;

var
  ln             : ^tln;
  s3minstpointer : array[0..99] of word;
  s3mpattpointer : array[0..255] of word;
  i,j            : integer;
  s3minstrument  : ts3minstr;
  p              : dataptr;
  patt           : tpatt;
  line,w,pos,ch  : word;
  data           : ^tba;
  s              : string;
  b              : byte;
  channeluse     : array[0..31] of boolean;
  lastporta      : array[0..31] of byte;
  lastvol        : array[0..31] of byte;

begin
  if (f.status <> stok) then fail;

  tsound.init(dest,0);

  { Check for SCRM id  }
  f.read(header,sizeof(header));
  if (makestring(header.id,4) <> 'SCRM') then
  begin
    tsound.done;
    fail;
  end;


  { Set song variables }
  songlength  := header.songlength;
  instruments := header.instruments;
  patterns    := header.patterns;

  { Load arrangement and pattern pointers }
  f.read(arrange,songlength);
  f.read(s3minstpointer,(header.instruments)*2);
  f.read(s3mpattpointer,header.patterns*2);

  { Check for Pan table and load if available }
  if (header.defpan = $FC) then f.read(panlist,32)
  else
    for i := 0 to 31 do panlist[i] := $20;

  t0volslide := (header.flags and $40 <> 0)or(header.version = $1300);

  arrangespeed[0].tempo := header.starttempo;
  arrangespeed[0].bpm   := header.startbpm;

  for i := 0 to 27 do
  if (header.songname[i] >= #32) then
    songname[i+1] := header.songname[i] else songname[i+1] := #32;
  songname[0] := #28;

  { Set format restrictions }
  format     := fmt_s3m;
  bpm        := header.startbpm;
  tempo      := header.starttempo;
  oscillator := s3m_freq;
  pershift   := 2;

  if (header.flags and $10 <> 0) then
  begin { Force Amiga limits }
    minslide   := 113*4;
    maxslide   := 856*4;
  end else
  begin
    minslide   := 56;
    maxslide   := 27392;
  end;

  { Filter out unused patterns }
{  writeln(patterns);
  for i := 0 to patterns-1 do if (s3mpattpointer[i] = 0) then break;
  if (i < patterns-1) then patterns := i;
  writeln(patterns);
  readln;}

  for i := 0 to songlength-1 do if (arrange[i] >= patterns)and(i < 255) then
  begin
    move(arrange[i+1],arrange[i],255-i);
    dec(songlength);
  end;
  if (songlength < 255) then inc(songlength);

  for i := 0 to songlength-1 do if (arrange[i] >= patterns) then break;
  songlength := i;

  { Check for any illegaly assigned patterns }
 { for i := 0 to songlength-1 do if (s3mpattpointer[arrange[i]] = 0) then fail;}

  getmem(data,32*6*64+6);

  { Filter out unused channels }
  for i := 0 to patterns-1 do if (s3mpattpointer[i] <> 0) then
  begin
    f.reset;
    f.seek(longint(s3mpattpointer[i])*16);
    f.read(w,2);
    if (w > 32*6*64) then
    begin
      tsound.done;
      fail;
    end;

    fillchar(data^,w,0);
    f.read(data^,w);
    if (f.status <> stok) then
    begin
      tsound.done;
      fail;
    end;

    line := 0;
    pos  := 0;

    while (line < 64) do
    begin
      if (pos < 32*6*64)and(data^[pos] <> 0) then
      begin
        b  := data^[pos];

        if (b and $20 <> 0) then
        begin
          if (data^[pos+2] <> 0) then
          channeluse[b and 31] := true;
          inc(pos,2);
        end;

        if (b and $40 <> 0) then inc(pos);
        if (b and $80 <> 0) then inc(pos,2);

      end else inc(line);
      inc(pos);
    end;
  end;

  playchnls := 0;
  fillchar(mapping,sizeof(mapping),$FF);
  for i := 0 to 31 do
  if (channeluse[i])and(header.channelinfo[i] and $80 = 0) then
  begin
    mapping[playchnls] := i;
    inc(playchnls);
  end;

  linesize    := playchnls*sizeof(tpatt);

  if (playchnls > 4) then tsound.setchannels(playchnls shr 1)
  else tsound.setchannels(playchnls);

  setbuffersize(round(dat.mainfreq/(bpm*2/5)));

  fillchar(channeluse,sizeof(channeluse),0);
  { Decompress and convert pattern data }
  for i := 0 to patterns-1 do
  begin
    with pattern[i] do
    begin
      size := linesize*64;
      rows := 63;
      getmem(p,size);
    end;
    fillchar(pattern[i].p^,pattern[i].size,0);
    if (s3mpattpointer[i] <> 0) then
    begin
      f.reset;
      f.seek(longint(s3mpattpointer[i])*16);
      f.read(w,2);
      if (w > 32*6*64) then
      begin
        tsound.done;
        fail;
      end;
      fillchar(data^,w,0);
      f.read(data^,w);
      if (f.status <> stok) then
      begin
        tsound.done;
        fail;
      end;

      line := 0;
      pos  := 0;

      while (line < 64) do
      begin
        if (pos < 32*6*64)and(data^[pos] <> 0) then
        begin
          b  := data^[pos];
          ch := mapping[b and 31];
          fillchar(patt,sizeof(patt),0);

          if (b and $20 <> 0) then
          begin
            case data^[pos+1] of
              255 : patt.per := 0;
              254 : patt.per := mdkeyoff;
              else patt.per := s3mtab[data^[pos+1] shr 4,data^[pos+1] and $F];
            end;
            patt.ins := data^[pos+2];
            inc(pos,2);
          end;

          if (b and $40 <> 0) then
          begin
            if (data^[pos+1] <= 64) then patt.vol := data^[pos+1]+$10;
            inc(pos);
          end;

          if (b and $80 <> 0) then
          begin
            patt.arg := data^[pos+2];
            case data^[pos+1] of
              $01 : patt.cmd := $F5; { S3M set tempo }
              $02 : patt.cmd := $0B;
              $03 : patt.cmd := $0D;
              $04 : begin { Volume slide }
                      if (patt.arg = 0) then
                      begin
                        data^[pos+2] := lastvol[ch];
                        patt.arg     := lastvol[ch];
                      end;
                      lastvol[ch] := patt.arg;
                      if ((patt.arg and $F = $F)and(patt.arg shr 4 <> 0))or
                         ((patt.arg and $F0 = $F0)and(patt.arg and $F <> 0)) then
                      begin
                        patt.cmd := $1E;
                        if (patt.arg and $F = $F) then
                          patt.arg := (patt.arg shr 4)or($A0) else
                          patt.arg := (patt.arg and $F)or($B0);
                      end
                        else patt.cmd := $0A;
                    end;
{              $04 : patt.cmd := $F6;}
              $05 : begin { Portamento down }
                      if (patt.arg = 0) then
                      begin
                        data^[pos+2] := lastporta[ch];
                        patt.arg     := lastporta[ch];
                      end;
                      lastporta[ch] := patt.arg;
                      if (patt.arg >= 0)and(patt.arg <= $DF) then
                        patt.cmd := $02;
                      if (patt.arg and $F0 = $F0) then
                      begin
                        patt.cmd := $1E;
                        patt.arg := (patt.arg and $F) or $20;
                      end;
                      if (patt.arg and $F0 = $E0) then
                      begin
                        patt.cmd := $21;
                        patt.arg := (patt.arg and $F) or $20;
                      end;
                    end;
              $06 : begin { Portamento up }
                      if (patt.arg = 0) then
                      begin
                        data^[pos+2] := lastporta[ch];
                        patt.arg     := lastporta[ch];
                      end;
                      lastporta[ch] := patt.arg;
                      if (patt.arg >= 0)and(patt.arg <= $DF) then
                        patt.cmd := $01;
                      if (patt.arg and $F0 = $F0) then
                      begin
                        patt.cmd := $1E;
                        patt.arg := (patt.arg and $F) or $10;
                      end;
                      if (patt.arg and $F0 = $E0) then
                      begin
                        patt.cmd := $21;
                        patt.arg := (patt.arg and $F) or $10;
                      end;
                    end;
              $07 : patt.cmd := $03; { Porta to note      }
              $08 : patt.cmd := $04; { Vibrato            }
              $09 : patt.cmd := $1D; { Tremor             }
              $0A : patt.cmd := $00; { Arpeggio           }
              $0B : patt.cmd := $06; { Vibr+vol slide     }
              $0C : patt.cmd := $05; { Porta+vol slide    }
              $0F : patt.cmd := $09; { Set sample ofs     }

              $11 : patt.cmd := $1B; { Retrig+vol slide   }
              $12 : patt.cmd := $07; { Tremolo            }
              $13 : patt.cmd := $0E;

              $14 : patt.cmd := $0F; { Set tempo/bpm      }
              $15 : patt.cmd := $F4; { Extra fine vibrato }
              $16 : patt.cmd := $10; { Set global volume  }
              $18 : patt.cmd := $08; { Panning            }
              else
              begin
                patt.cmd := 0;
                patt.arg := 0;
              end;
            end;
            if (patt.cmd = $0E) then
            begin
              case (patt.arg shr 4) of
                $0 : if (patt.arg <> 1)and(patt.arg <> 0) then
                     begin;patt.arg := 0;patt.cmd := 0;end;
                $1 : patt.arg := (patt.arg and $F) + $30; { Set glissando        }
                $2 : patt.arg := (patt.arg and $F) + $50; { Set finetune         }
                $3 : patt.arg := (patt.arg and $F) + $40; { Set vibrato waveform }
                $4 : patt.arg := (patt.arg and $F) + $70; { Set tremolo waveform }

                $A : begin { Stereo control - old panning }
                       if (patt.arg < 8) then inc(patt.arg,8) else
                       dec(patt.arg,8);

                       patt.arg := (patt.arg and $F) or $80;
                       patt.cmd := $0E;
                    end;

                $B : patt.arg := (patt.arg and $F) + $60;
              end;
            end;
            if (patt.cmd = $1E) then patt.cmd := $0E;
            inc(pos,2);
          end;
          if (ch < playchnls) then
          {$IFNDEF FPC}
          move(patt,ptr(seg(pattern[i].p^),ofs(pattern[i].p^)+line*linesize+ch*sizeof(patt))^,sizeof(patt));
          {$ELSE}
          move(patt,pointer(pattern[i].p+line*linesize+ch*sizeof(patt))^,sizeof(patt));
          {$ENDIF}
        end else inc(line);
        inc(pos);
      end;
    end;
  end;
  freemem(data,32*6*64+6);

  cs.init;
  cs.globalvol := header.globalvol*4;

  if (header.mastervol and $80 = 0)and(header.defpan <> $FC) then { Mono }
  for i := 0 to 31 do cs.channel[i].panning := 0
    else
  for i := 0 to 31 do if (mapping[i] < playchnls) then with cs.channel[mapping[i]] do
  begin
    if (panlist[i] and $20 <> 0)or(panlist[i] = 0) then
    begin
      panning := (i and 1)*512-256;
      if panning < 0 then
        inc(panning,51) else
        dec(panning,51);
    end
    else panning := integer(panlist[i])*32-256;
  end;

  for i := 0 to playchnls-1 do cs.channel[i].sound.init(self);


  { Load instrument samples }
  for i := 1 to instruments do
  begin
    f.reset;
    f.seek(longint(s3minstpointer[i-1])*16);
    f.read(s3minstrument,sizeof(ts3minstr));
    if (f.status = stok) then

    with cs.instrument[i] do
    begin
      f.reset;
      f.seek(longint(s3minstrument.segment)*16);
      { Ignore, if sample not included }
      if (f.status <> stok) then continue;

      size  := s3minstrument.size;
      name  := s3minstrument.samplename;
      ofreq := stdfreq;
      tuning := s3minstrument.c2spd;
      if (s3minstrument.flags and 1 <> 0) then
      begin
        loopstart := s3minstrument.loopstart;
        loopend   := s3minstrument.loopend;
      end else
      begin
        loopstart := 0;
        loopend   := 0;
      end;
      volume  := s3minstrument.volume;
      volfade := 65536;

      if (cs.instrument[i].size > 0) then
      with cs.instrument[i] do
      begin
        if (s3minstrument.flags and $4 = 0) then { 8 bit data }
        begin
          {$IFNDEF FPC}
          getmem32(s,size shl 1);
          getmem32(p,size);
          f.read32(ptr48(p,0),size);
          {$ELSE}
          getmem(s,size shl 1);
          getmem(p,size);
          f.read(p^,size);
          {$ENDIF}
          if (f.status <> stok) then
          begin
            {$IFNDEF FPC}
            freemem32(s);
            freemem32(p);
            {$ELSE}
            freemem(s);
            freemem(p);
            {$ENDIF}
            size := 0;
            open := false;
            continue;
          end
          else open := true;
          {$IFNDEF FPC}
          aud_us8_to_s16(ptr48(p,0),ptr48(s,0),size);
          freemem32(p);
          {$ELSE}
          aud_us8_to_s16(p,s,size);
          freemem(p);
          {$ENDIF}
        end else
        begin
          {$IFNDEF FPC}
          getmem32(s,size);
          f.read32(ptr48(s,0),size);
          aud_us16_to_s16(ptr48(s,0),ptr48(s,0),size shr 1);
          {$ELSE}
          getmem(s,size);
          f.read(s^,size);
          aud_us16_to_s16(s,s,size shr 1);
          {$ENDIF}
          if (f.status <> stok) then
          begin
            {$IFNDEF FPC}
            freemem32(s);
            {$ELSE}
            freemem(s);
            {$ENDIF}
            size := 0;
            open := false;
            continue;
          end
          else open := true;
          size      := size shr 1;
          loopstart := loopstart shr 1;
          loopend   := loopend shr 1;
        end;
      end;
    end;
  end;
  cs.ramp; { Apply volume ramping }
  initialized := true;
end;

procedure ts3m.play;
var
  i : integer;

begin
  if (flags and snd_pause <> 0) then
  begin
    flags := (flags and not snd_pause) or snd_play;
    for i := 0 to playchnls-1 do cs.channel[i].sound.pause;
  end
  else
  begin
    actline    := 0;
    actpattern := 0;
    flags      := flags or snd_play;
    tick       := 0;
    tempo      := header.starttempo;
    bpm        := header.startbpm;
  end;
end;

procedure decomp8(data : pointr;size : longint);
var
  old,nw : shortint;
  i      : longint;

begin
  old := 0;

  {$IFNDEF FPC}
  for i := 0 to size-1 do
  begin
    nw := shortint(readmem32b(data))+old;
    writemem32b(data,nw);
    old := nw;
    inc(tselofs32(data).offset32);
  end;
  {$ELSE}
  for i := 0 to size-1 do
  begin
    nw := shortint(data^)+old;
    byte(data^) := nw;
    old := nw;
    inc(data);
  end;
  {$ENDIF}
end;

procedure decomp16(data : pointr;size : longint);
var
  old,nw : integer;
  i       : longint;

begin
  old := 0;

  {$IFNDEF FPC}
  for i := 0 to size-1 do
  begin
    nw := integer(readmem32w(data))+old;
    writemem32w(data,nw);
    old := nw;
    inc(tselofs32(data).offset32,2);
  end;
  {$ELSE}
  for i := 0 to size-1 do
  begin
    nw := integer(data^)+old;
    word(data^) := nw;
    old := nw;
    inc(data,2);
  end;
  {$ENDIF}
end;

constructor txm.init(var dest : tsounddriver;var f : tstream);
type
  ptinfo = packed record
    headersize : longint;
    pack       : byte;
    rows       : word;
    size       : word;
  end;

  tln = array[0..31] of tpatt;

  txmpatt = packed record
    note : byte;
    ins  : byte;
    vol  : byte;
    cmd  : byte;
    arg  : byte;
  end;

  txminshdr = packed record
    size       : longint;
    samplename : array[0..21] of char;
    itype      : byte;
    samples    : word;
  end;

  txmins2hdr = packed record
    headersize : longint;
    smpmap     : array[0..95] of byte;
    venvelope  : array[0..11] of record
                   x,y : word;
                 end;
    penvelope  : array[0..11] of record
                   x,y : word;
                 end;
    vpoints    : byte;
    ppoints    : byte;
    vsustain   : byte;
    vlstart    : byte;
    vlend      : byte;
    pustain    : byte;
    plstart    : byte;
    plend      : byte;
    vtype      : byte;
    ptype      : byte;
    vitype     : byte;
    visweep    : byte;
    videpth    : byte;
    virate     : byte;
    vfadeout   : word;
    reserved   : array[0..10] of byte;
  end;

  txmsamplehdr = packed record
    size       : longint;
    loopstart  : longint;
    looplength : longint;
    volume     : byte;
    finetune   : shortint;
    flags      : byte;
    panning    : byte;
    relnote    : shortint;
    reserved   : byte;
    samplename : array[0..21] of char;
  end;

var
  i,i2,j  : integer;
  fpos    : longint;
  info    : ptinfo;
  datap   : pointer;
  b       : byte;
  patt    : tpatt;
  pos,ch  : word;
  line    : word;
  ihdr1   : txminshdr;
  ihdr2   : txmins2hdr;
  p       : dataptr;
  np      : longint;
  count   : longint;
  sample  : array[0..15] of record
              smp    : txmsamplehdr;
            end;
  currsample : word;
  x,memsize  : longint;


{$IFDEF DEBUG}
function inttostr(l : longint) : string;
{ convert longint to string }
var
  c : string[20];
begin
  str(l,c);
  inttostr := c;
end;

procedure xmerror(s : string;expect,read : longint);
begin
  writeln;
  writeln('TXM_INIT: '+s);
  writeln('Stream error  : ',f.status);
  writeln('I/O result    : ',f.errorinfo);
  writeln('File position : ',f.getpos);
  writeln('Data expected : ',expect);
  writeln('Data read     : ',count);
end;
{$ENDIF}

function getbyte : byte;
type
  tba = array[0..0] of byte;
begin
  {$IFNDEF FPC}
  getbyte := tba(datap^)[pos];
  {$ELSE}
  getbyte := byte(pointer(datap+pos)^);
  {$ENDIF}
  inc(pos);
end;


begin
  {$IFDEF DEBUG}
  writeln('TXM_INIT: Debug mode on');
  {$ENDIF}
  if (f.status <> stok) then
  begin
    {$IFDEF DEBUG}
    writeln('TXM_INIT: File error');
    {$ENDIF}
    fail;
  end;

  tsound.init(dest,0);

  f.read(startinfo,sizeof(startinfo));
  if (makestring(startinfo.id,17) <> 'Extended Module: ') then
  begin
    {$IFDEF DEBUG}
    writeln('TXM_INIT: wrong ID "'+makestring(startinfo.id,17)+'"');
    {$ENDIF}
    tsound.done;
    fail;
  end;

  fpos := f.getpos;
  f.read(header,sizeof(header));
  f.read(arrange,256);
  f.seek(header.headersize+fpos);

  for i := 0 to 19 do
  if (startinfo.songname[i] >= #32) then
    songname[i+1] := startinfo.songname[i] else songname[i+1] := #32;
  songname[0] := #20;

  {$IFDEF DEBUG}
  writeln('TXM_INIT: Loading module : "'+songname+'"');
  writeln;
  {$ENDIF}

  songlength  := header.songlength;

  instruments := header.instruments;

  patterns    := header.patterns;
  playchnls   := header.channels;
  linesize    := playchnls*sizeof(tpatt);

  format      := fmt_xm;
  oscillator  := xm_freq;
  pershift    := 0;
  minslide    := 457;
  maxslide    := 907;
  bpm         := header.startbpm;
  tempo       := header.starttempo;
  arrangespeed[0].tempo := header.starttempo;
  arrangespeed[0].bpm   := header.startbpm;
  getmem(smpmapping,sizeof(txmsmpmapping));

  if (playchnls > 4) then tsound.setchannels(playchnls shr 1)
  else tsound.setchannels(playchnls);

  setbuffersize(round(dat.mainfreq/(bpm*2/5)));

  cs.init;

  for i := 0 to playchnls-1 do cs.channel[i].sound.init(self);

  {$IFDEF DEBUG}
  writeln('-- Loading patterns --');
  writeln('Pattern count: ',patterns);
  writeln;
  {$ENDIF}
  getmem(datap,playchnls*5*256);
  for i := 0 to patterns-1 do
  begin
    f.reset;
    np := f.getpos;
    f.read(info,sizeof(info));
    inc(np,info.headersize);
    f.seek(np);
    with pattern[i] do
    begin
      size := linesize*info.rows;
      rows := info.rows-1;
      getmem(p,size);
    end;
    { If pattern empty, create empty pattern }
    if (info.size = 0) then with pattern[i] do
    begin
      fillchar(p^,size,0);
      {$IFDEF DEBUG}
      writeln('Pattern ',i,' empty    ');
      {$ENDIF}
    end
    else
    begin { Else decompress pattern }
      {$IFDEF DEBUG}
      write('Loading pattern ',i,#13);
      {$ENDIF}
      count := f.getpos;
      f.read(datap^,info.size);
      count := f.getpos-count;
      if (f.status <> stok) then
      begin
        {$IFDEF DEBUG}
        xmerror('Pattern load failed',info.size,count);
        {$ENDIF}
        tsound.done;
        fail;
      end;

      pos := 0;
      for line := 0 to info.rows-1 do
      for ch := 0 to playchnls-1 do if (pos < info.size) then
      begin
        b := getbyte;
        if (b and $80 <> 0) then
        begin
          fillchar(patt,sizeof(patt),0);
          { Packed note       }
          if (b and $01 <> 0) then
          begin
            patt.per := getbyte;
            if (patt.per = 97) then patt.per := mdkeyoff
            else if (patt.per > 96) then patt.per := 96;
          end;
          { Packed instrument }
          if (b and $02 <> 0) then patt.ins := getbyte;
          { Packed volume     }
          if (b and $04 <> 0) then patt.vol := getbyte;
          { Packed command    }
          if (b and $08 <> 0) then patt.cmd := getbyte;
          { Packed argument   }
          if (b and $10 <> 0) then patt.arg := getbyte;
        end else { No packed data }
        begin
          patt.per := b and $7F;
          if (patt.per = 97) then patt.per := mdkeyoff
          else if (patt.per > 96) then patt.per := 96;
          patt.ins := getbyte;
          patt.vol := getbyte;
          patt.cmd := getbyte;
          patt.arg := getbyte;
        end;
        {$IFNDEF FPC}
        move(patt,ptr(seg(pattern[i].p^),ofs(pattern[i].p^)+line*linesize+ch*sizeof(patt))^,sizeof(patt));
        {$ELSE}
        move(patt,pointer(pattern[i].p+line*linesize+ch*sizeof(patt))^,sizeof(patt));
        {$ENDIF}
      end;
      if (pos <> info.size) then
      begin
        {$IFDEF DEBUG}
        writeln;
        writeln('TXM_INIT: Compressed data size difference');
        writeln('Bytes loaded      : ',info.size);
        writeln('Bytes decompressed: ',pos);
        {$ENDIF}
        tsound.done;
        fail;
      end;
    end;
  end;
  freemem(datap,playchnls*5*256);
  {$IFDEF DEBUG}
  writeln('TXM_INIT: Patterns successfully decompressed');
  writeln('File position     : ',f.getpos);
  writeln('Left bytes in file: ',f.getsize-f.getpos);
  writeln;
  {$ENDIF}

  {$IFDEF DEBUG}
  writeln('-- Loading instruments --');
  writeln('Instrument count: ',header.instruments);
  writeln;
  {$ENDIF}
  { Load all instruments }
  currsample := 1;
  for i := 0 to header.instruments-1 do
  begin
    np := f.getpos;
    f.read(ihdr1,sizeof(ihdr1));
    inc(np,ihdr1.size);
    if (f.status <> stok) then
    begin
      {$IFDEF DEBUG}
      xmerror('instrument #'+inttostr(i)+' header load failed',sizeof(txmsamplehdr),count);
      {$ENDIF}
      tsound.done;
      fail;
    end;

    if (ihdr1.samples > 0) then
    begin
      f.reset;
      f.read(ihdr2,sizeof(ihdr2));
      f.seek(np);
      move(ihdr2.smpmap,smpmapping^[i+1],96);

      if (ihdr1.samples > 15) then
      begin
        tsound.done;
        fail;
      end;
      for j := 0 to ihdr1.samples-1 do
      begin
{        count := f.getpos;}
        f.read(sample[j].smp,sizeof(sample[j].smp));
{        count := f.getpos-count;}
      end;

      for j := 0 to ihdr1.samples-1 do with cs.instrument[currsample] do
      with sample[j] do
      begin
        if (smp.size = 0)and(j > 0) then continue;
        mapping[i+1,j] := currsample;
        inc(currsample);
        {$IFDEF DEBUG}
        write('Loading instrument ',i,', Sample ',j,#13);
        {$ENDIF}
        { Restart if sample empty }

        { Convert sample name }
        for i2 := 0 to 21 do
        if (ihdr1.samplename[i2] >= #32) then
          name[i2+1] := ihdr1.samplename[i2] else name[i2+1] := #32;
        name[0] := #22;
        if (smp.size = 0) then continue;

        { Set looping
          0 : no looping
          1 : normal looping
          2 : ping-pong
        }

        { Set sample variables }
        volume  := smp.volume;
        relnote := smp.relnote;
        panning := (integer(smp.panning)-128)*2;
        ofreq   := 8363;
        tuning  := smp.finetune;
        size    := smp.size;
        volfade := ihdr2.vfadeout;
        if (volfade = 0) then volfade := 65536;

        memsize := size;
        if (smp.flags and 3 <> 0) then
        begin
          if (smp.loopstart+smp.looplength*2 > size) then
          memsize := smp.loopstart+smp.looplength*2;
        end;

        if (memsize > memavail) then
        begin
          tsound.done;
          fail;
        end;
        { Load and convert sample data }
        if (smp.size > 0) then
        begin
          if (smp.flags and $10 = 0) then { 8 bit data }
          begin
            {$IFNDEF FPC}
            getmem32(s,memsize shl 1);
            getmem32(p,size);
            count := f.read32(ptr48(p,0),size);
            {$ELSE}
            getmem(s,memsize shl 1);
            getmem(p,size);
            count := f.getpos;
            f.read(p^,size);
            count := f.getpos-count;
            {$ENDIF}
            if (f.status <> stok) then
            begin
              {$IFNDEF FPC}
              freemem32(s);
              freemem32(p);
              {$ELSE}
              freemem(s);
              freemem(p);
              {$ENDIF}
              {$IFDEF DEBUG}
              xmerror('Sample data load failed, instrument #'+inttostr(i)+' Sample #'+inttostr(j),size,count);
              {$ENDIF}
              size := 0;
              open := false;
              {continue;}
              tsound.done;
              fail;
            end
            else open := true;
            {$IFNDEF FPC}
            decomp8(ptr48(p,0),size);
            aud_s8_to_s16(ptr48(p,0),ptr48(s,0),size);
            freemem32(p);
            {$ELSE}
            decomp8(p,size);
            aud_s8_to_s16(p,s,size);
            freemem(p);
            {$ENDIF}
          end else
          begin  { 16 bit data }
            {$IFNDEF FPC}
            getmem32(s,memsize);
            f.read32(ptr48(s,0),size);
            {$ELSE}
            getmem(s,memsize);
            f.read(s^,size);
            {$ENDIF}
            if (f.status <> stok) then
            begin
              {$IFNDEF FPC}
              freemem32(s);
              {$ELSE}
              freemem(s);
              {$ENDIF}
              size := 0;
              open := false;
              {continue;}
              tsound.done;
              fail;
            end
              else open := true;
            size := size shr 1;
            memsize := memsize shr 1;
            smp.loopstart  := smp.loopstart shr 1;
            smp.looplength := smp.looplength shr 1;
            {$IFNDEF FPC}
            decomp16(ptr48(s,0),size);
            {$ELSE}
            decomp16(s,size);
            {$ENDIF}
          end;
        end; { Load sample data }
        if (smp.flags and 3 <> 0)and(smp.looplength > 0) then
        begin
          loopstart := smp.loopstart;
          loopend   := smp.loopstart+smp.looplength;
          if (smp.flags and 3 = 2) then
          begin { Apply pingpong-loop }
            {$IFNDEF FPC}
            for x := 0 to smp.looplength-1 do
            writemem32w(ptr48(s,(loopend+smp.looplength-x-1)*2),
                        readmem32w(ptr48(s,(x+loopstart)*2)));
            {$ELSE}
            for x := 0 to smp.looplength-1 do
            word(pointer(s+(loopend+smp.looplength-x-1)*2)^) :=
                        word(pointer(s+(x+loopstart)*2)^);
            {$ENDIF}
            inc(loopend,smp.looplength);
            size := memsize;
          end;
        end else
        begin
          loopstart := 0;
          loopend   := 0;
        end;

      end; { Load samples }
    end else { Empty instrument }
    begin
      inc(currsample);
      f.seek(np);
    end;
  end; { Load instruments }
  instruments := currsample-1;
  cs.ramp; { Apply volume ramping }
  initialized := true;
  {$IFDEF DEBUG}
  writeln('TXM_INIT: Loading successful        ');
  writeln('Press enter to continue...');
  readln;
  {$ENDIF}
end;

destructor txm.done;
begin
  inherited done;
  freemem(smpmapping,sizeof(txmsmpmapping));
end;

procedure txm.play;
var
  i : integer;

begin
  if (flags and snd_pause <> 0) then
  begin
    flags := (flags and not snd_pause) or snd_play;
    for i := 0 to playchnls-1 do cs.channel[i].sound.pause;
  end
  else
  begin
    actline    := 0;
    actpattern := 0;
    flags      := flags or snd_play;
    tick       := 0;
    tempo      := header.starttempo;
    bpm        := header.startbpm;
  end;
end;

procedure txm.calcline;
type
  tln = array[0..31] of tpatt;

var
  ln        : ^tln;
  w         : word;
  ch,i      : integer;
  dontplay  : boolean;
  curr      : tpatt;
  per       : longint;
  last      : word;
  broken    : boolean;

begin
  if (actline = word(pattern[arrange[actpattern]].rows)+1) then
  begin
    actline := 0;
    inc(actpattern);
    arrangespeed[actpattern].bpm   := bpm;
    arrangespeed[actpattern].tempo := tempo;
  end;

  if (actpattern >= songlength) then
  begin
    flags := flags and not snd_play;
    exit;
  end;

  {$IFNDEF FPC}
  ln := ptr(seg(pattern[arrange[actpattern]].p^),
            ofs(pattern[arrange[actpattern]].p^)+actline*linesize);
  {$ELSE}
  ln := pointer(pattern[arrange[actpattern]].p+actline*linesize);
  {$ENDIF}

  broken := false; { Clear pattern not broken flag }

  for ch := 0 to playchnls-1 do with cs.channel[ch] do
  begin
    curr := ln^[ch];
    if (curr.ins <> 0)and(curr.per < 97)and(curr.per > 0) then
      curr.ins := mapping[curr.ins,smpmapping^[curr.ins,curr.per]]
      else curr.ins := 0;
    per  := curr.per;

    { Ignore porta to note if instrument different }
    if (curr.cmd = $03)and(curr.ins <> 0)and(curr.ins <> voice) then
    begin
      curr.cmd := 0;
      curr.arg := 0;
    end;

    if (per <> 0)and(per <> mdkeyoff) then
    begin
      w := curr.ins;
      inc(per,cs.instrument[w].relnote);
      if (w = 0) then w := voice;

      i := (per mod 12)*8 + cs.instrument[w].tuning div 16;
      if (i < 0) then i := 0 else if (i > 102) then i := 102;
      per := round((xmtab[i]*(1-frac(cs.instrument[w].tuning/16))+
                    xmtab[i+1]*(frac(cs.instrument[w].tuning/16)))
                    *16/exp((per div 12)*system.ln(2)));

{     per := (7680 - (per * 64) - (cs.instrument[w].tuning div 2));
     freq := round(8363*exp((4608 - per)*system.ln(2)/768));}

      if (per = 0) then per := 1;
      if (curr.cmd <> $03) then
      begin
        period := per;
        freq   := oscillator div per;
      end;
    end;

    { Don't play note if note delay, porta to note or key off }
    dontplay := ((curr.cmd = $0E)and((curr.arg shr 4) = $0D))or
                 ((curr.cmd = $03){and((curr.ins = 0)or(curr.ins = voice))})or(curr.per = mdkeyoff);

    if (not dontplay) then
    if (curr.ins <> 0) then
    begin
      panning := cs.instrument[curr.ins].panning;
      cs.play(ch,curr.ins,freq,cs.instrument[curr.ins].volume);
    end
    else if (per <> 0)or((curr.ins <> 0)and(per = 0)) then
    begin
      panning := cs.instrument[curr.ins].panning;
      cs.play(ch,voice,freq,volume);
    end;

    { Retrigger volume if porta note and same instrument on line }
    if (curr.cmd = $03)and(curr.ins <> 0)and(curr.ins = voice)and
       (format <> fmt_mod)and(curr.vol = 0)
       then curr.vol := cs.instrument[voice].volume+$10;

    if ((arpeggio)or(vibrato))and(period <> 0) then
    begin
      cs.setfrequency(ch,freq);
      arpeggio    := false;
      vibrato     := false;
    end;

    { Reset all FX which are linewise }
    tremolo     := false;
    porta       := false;
    bfinepslide := false;
    bfinevslide := false;
    vslide      := false;
    bportaslide := false;
    retrig      := 0;
    tremor      := false;
    last        := lasteffect;
    lasteffect  := curr.cmd+curr.arg shl 8;

    if (curr.vol <> 0) then
    begin
      if (curr.vol <= $50) then
      begin
        volume := curr.vol-$10;
        if ((curr.cmd = $0E)and((curr.arg shr 4) = $0D))
        then else cs.setvolume(ch,volume);
      end else
      begin
         { XM FX... not complete, haven't had good docu }
         case (curr.vol shr 4) of
           $6 : begin;volumeslide := -(curr.vol and $F);vslide := true;end;
           $7 : begin;volumeslide := (curr.vol and $F);vslide := true;end;
           $C : begin
                  panning := (curr.vol and $F);
                  if (panning > 7) then inc(panning);
                  panning := panning*32-256;
                  cs.setvolume(ch,cs.channel[ch].volume);
                end;
           $D : begin
                  if (panning-curr.vol and $F > -256) then
                    dec(panning,curr.vol and $F) else panning := -256;
                  cs.setvolume(ch,cs.channel[ch].volume);
                end;
           $E : begin
                  if (panning+curr.vol and $F < 256) then
                    inc(panning,curr.vol and $F) else panning := 256;
                  cs.setvolume(ch,cs.channel[ch].volume);
                end;
         end;
      end;
    end;

    { Check for key off }
    if (curr.per = mdkeyoff) then
    begin
      {volume := 0;
      cs.setvolume(ch,volume);}
      vfading := true;
    end;

    { Process effects }
    case curr.cmd of
      $00 : if (curr.arg <> 0) then
            begin{ok; Arpeggio          }
              arpeggio  := true;
              arpeggiox := (curr.arg shr 4);
              arpeggioy := (curr.arg and $F);
            end else lasteffect := mdkeyoff;
      $01 : begin{ok; Portamento down   }
              if (curr.arg <> 0) then
                portaslide  := (integer(curr.arg) shl pershift);
              if (portaslide > 0) then portaslide := -portaslide;
              bportaslide := (portaslide <> 0);
            end;
      $02 : begin{ok; Portamento up     }
              if (curr.arg <> 0) then
                portaslide  := integer(curr.arg) shl pershift;
              if (portaslide < 0) then portaslide := -portaslide;
              bportaslide := (portaslide <> 0);
            end;
      $03 : begin{ok; Porta to note     }
              if (per <> 0) then portanote := per;
              if (curr.arg <> 0) then
              begin
                portaspeed := word(curr.arg) shl pershift;
                if (period > portanote) then portaspeed := -portaspeed;
              end;
              porta := (portanote <> 0);
            end;
      $04 : begin{ok; Set vibrato       }
              if (curr.arg and $F0 <> 0) then vibratospeed := curr.arg shr 4;
              if (curr.arg and $F <> 0) then vibratodepth := (curr.arg and $F) shl pershift;
              vibrato := true;
            end;
      $05 : begin{ok; Porta+vol slide   }
              if (curr.arg and $F <> 0) then volumeslide := -(curr.arg and $F)
                else if (curr.arg and $F0 <> 0) then volumeslide := curr.arg shr 4;
              porta := (portanote <> 0);
              vslide := true;
            end;
      $06 : begin{ok; Vibrato+vol slide }
              if (curr.arg and $F <> 0) then volumeslide := -(curr.arg and $F)
              else if (curr.arg and $F0 <> 0) then volumeslide := curr.arg shr 4;
              vibrato := true;
              vslide  := true;
            end;
      $07 : begin{ok; Tremelo           }
              if (curr.arg and $F0 <> 0) then tremolospeed := curr.arg shr 4;
              if (curr.arg and $F <> 0) then tremolodepth := curr.arg and $F;
              tremolo := true;
            end;
      $08 : begin{ok; Panning           }
              panning := integer(curr.arg)*2-256;
              cs.setvolume(ch,cs.channel[ch].volume);
            end;
      $09 : begin{ok; Set instrument pos}
              if (curr.arg <> 0) then lastsetpos := longint(curr.arg) shl 8;
              cs.setpos(ch,lastsetpos);
            end;
      $0A : begin
              if (curr.arg and $F <> 0) then volumeslide := -(curr.arg and $F)
              else if (curr.arg and $F0 <> 0) then volumeslide := curr.arg shr 4;
              vslide  := (volumeslide <> 0);
            end;
      $0B : if (curr.arg <= patterns) then
            begin{ok}
              actline    := 63;
              actpattern := curr.arg-1;
            end;
      $0C : begin{ok; Set channel volume }
              if (curr.arg > 64) then curr.arg := 64;
              cs.channel[ch].volume := curr.arg;
              cs.setvolume(ch,curr.arg);
            end;
      $0D : if (not broken) then
            begin{ok; Pattern break      }
              inc(actpattern);
              actline := 10*(curr.arg shr 4)+curr.arg and $0F-1;
              if (actline > 62) then actline := 62;
              broken := true;
            end;
      $0E : case (curr.arg shr 4) of
              $01 : begin { ok; Fine portamento up   }
                      if (curr.arg and $F <> 0) then
                        finepslide := -((curr.arg and $F) shl pershift);
                      bfinepslide := (finepslide <> 0);
                    end;
              $02 : begin { ok; Fine portamento down }
                      if (curr.arg and $F <> 0) then
                        finepslide := (curr.arg and $F) shl pershift;
                      bfinepslide := (finepslide <> 0);
                    end;
              $04 : begin { Set vibrato waveform }
                      vibratotype := curr.arg and 7;
                      if (vibratotype and 3) = 3 then
                        vibratotype := random(2)+curr.arg and 4;
                    end;
              $05 : begin { Set finetune }
                      cs.instrument[voice].tuning := getc2spd(curr.arg and $F);
                    end;
              $07 : begin { Set tremolo waveform }
                      tremolotype := curr.arg and 7;
                      if (tremolotype and 3) = 3 then
                        tremolotype := random(2)+curr.arg and 4;
                    end;
              $08 : begin { ok; Position panning }
                      panning := (curr.arg and $F);
                      if (panning > 7) then inc(panning);
                      panning := panning*32-256;
                      cs.setvolume(ch,cs.channel[ch].volume);
                    end;
              $09 : begin { ok; Retrigger note   }
                      retrig     := (curr.arg and $F);
                      retrigtype := 0;
                    end;
              $0A : begin { ok; Fine volume slide up   }
                      if (curr.arg and $F <> 0) then
                        finevslide := curr.arg and $F;
                      bfinevslide := (finevslide <> 0);
                    end;
              $0B : begin { ok; Fine volume slide down }
                      if (curr.arg and $F <> 0) then
                        finevslide := -integer(curr.arg and $F);
                      bfinevslide := (finevslide <> 0);
                    end;
              $0C : begin { ok; Cut note }
                      if (curr.arg and $F <> 0) then volcutcount := (curr.arg and $F);
                    end;
              $0D : begin { ok; Note delay }
                      notedelay := (curr.arg and $F);
                      if (curr.ins <> 0) then voice := curr.ins;
                      if (curr.vol = 0) then volume := cs.instrument[voice].volume;
                    end;
              $0E : begin { ok; Pattern delay }
                      patdelay := (curr.arg and $F);
                    end;
            end;
      $0F : if (curr.arg <> 0) then
            begin{ok}
              if (curr.arg < 32) then tempo := curr.arg
              else
              begin
                bpm := curr.arg;
                setbuffersize(round(dat.mainfreq/(bpm*2/5)));
              end;
            end;
      $10 : begin { Set global volume }
              if (curr.arg > 64) then curr.arg := 64;
              cs.globalvol := curr.arg*4;
              for i := 0 to playchnls-1 do cs.setvolume(i,cs.channel[i].volume);
            end;
      { $11 : global vol slide  }
      { $15 : set envelope pos  }
      { $19 : panning slide pos }
      $1B : begin
              retrig     := (curr.arg and $F);
              retrigtype := (curr.arg shr 4);
            end;
      $1D : begin { Tremor }
              tremor      := true;
              tremorcount := 0;
              tremorx     := curr.arg shr 4+1;
              tremory     := tremorx+curr.arg and 7+1;
            end;

      $21 : case (curr.arg shr 4) of
              $01 :  begin { ok; Extra fine portamento up   }
                      if (curr.arg and $F <> 0) then
                        finepslide := -(curr.arg and $F);
                      bfinepslide := (finepslide <> 0);
                    end;
              $02 : begin { ok; Extra Fine portamento down }
                      if (curr.arg and $F <> 0) then
                        finepslide := curr.arg and $F;
                      bfinepslide := (finepslide <> 0);
                    end;
            end;
      else lasteffect := mdnotavail;
    end;
  end;
  inc(actline);
end;

begin
end.
